﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Data;
using System.Data.Linq;
using System.Data.Linq.Mapping;
using System.Transactions;
using System.Reflection;
using System.Runtime.Serialization;
using System.Text;
using System.Xml;
using System.IO;

namespace DevLeap.Linq.LinqToSql {
    class UpdateDemo {
        public static void SimpleDataChange() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                // 
                // Listing 6-3
                // 
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var customer = db.Customers.Single(c => c.CustomerID == "FRANS");
                Console.WriteLine(customer.ContactName); // "Paolo Accorti"
                customer.ContactName = "Marco Russo";
                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void SimpleInsertDirect() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var newCustomer = new Customer { 
                                        CustomerID = "DLEAP", 
                                        CompanyName = "DevLeap", 
                                        Country = "Italy" };
                db.Customers.InsertOnSubmit(newCustomer);
                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void SimpleRemove() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                // The following commented statements are the path to get 
                // the first order detail of the first order for the FRANS customer
                // We wrote a shorter version for the sake of simplicity in the book,
                // using directly an OrderID and a ProductID to identify the order detail row
                //var customer = db.Customers.Single(c => c.CustomerID == "FRANS");
                //var order = customer.Orders.First();
                //var oldDetail = order.Order_Details.First();
                var oldDetail = db.Order_Details.Single(od => od.OrderID == 10422 && od.ProductID == 26);

                db.Order_Details.DeleteOnSubmit(oldDetail);
                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void CascadingRemove() {
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var oldCustomer = db.Customers.Single(c => c.CustomerID == "FRANS");
                foreach (var order in oldCustomer.Orders) {
                    db.Order_Details.DeleteAllOnSubmit(order.Order_Details);
                }
                //db.Orders.DeleteAllOnSubmit(oldCustomer.Orders);
                db.Customers.DeleteOnSubmit(oldCustomer);
                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void SimpleInsertIndirect() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var newCustomer = new Customer { CustomerID = "DLEAP", CompanyName = "DevLeap", Country = "Italy" };
                var order = db.Orders.Single(o => o.OrderID == 10248);
                order.Customer = newCustomer;
                foreach (var o in newCustomer.Orders) {
                    Console.WriteLine("{0}-{1}", o.CustomerID, o.OrderID);
                }

                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void MoveDataBetweenCollections() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var oldCustomer = db.Customers.Single(c => c.CustomerID == "VINET");
                var newCustomer = new Customer { 
                                    CustomerID = "DLEAP", 
                                    CompanyName = "DevLeap", 
                                    Country = "Italy" };
                db.Customers.InsertOnSubmit(newCustomer);
                var order = oldCustomer.Orders.Single(o => o.OrderID == 10248);
                oldCustomer.Orders.Remove(order);
                newCustomer.Orders.Add(order);

                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void TreeRemove() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var order = db.Orders.Single(o => o.OrderID == 10248);
                db.Orders.DeleteOnSubmit(order);
                db.Order_Details.DeleteAllOnSubmit(order.Order_Details);

                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void ChangePrimaryKey() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;
                var oldCustomer = db.Customers.Single(c => c.CustomerID == "FRANS");
                Customer newCustomer = new Customer();
                newCustomer.CustomerID = "CHNGE";
                newCustomer.Address = oldCustomer.Address;
                newCustomer.City = oldCustomer.City;
                newCustomer.CompanyName = oldCustomer.CompanyName;
                newCustomer.ContactName = oldCustomer.ContactName;
                newCustomer.ContactTitle = oldCustomer.ContactTitle;
                newCustomer.Country = oldCustomer.Country;
                newCustomer.Fax = oldCustomer.Fax;
                newCustomer.Orders = oldCustomer.Orders;
                newCustomer.Phone = oldCustomer.Phone;
                newCustomer.PostalCode = oldCustomer.PostalCode;
                newCustomer.Region = oldCustomer.Region;

                Helper.DumpChanges(db.GetChangeSet());
                db.SubmitChanges();
            }
        }
        public static void CircularReference() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new CustomNorthwind(Connections.ConnectionString);
                db.Log = Console.Out;
                Employee empMarco = new Employee();
                db.Employees.InsertOnSubmit(empMarco);
                empMarco.FirstName = "Marco";
                empMarco.LastName = "Russo";
                Employee empPaolo = new Employee();
                empPaolo.FirstName = "Paolo";
                empPaolo.LastName = "Pialorsi";
                empPaolo.Employees.Add(empMarco);

                db.SubmitChanges();

                Console.WriteLine(Helper.Dump(empMarco));
                Console.WriteLine(Helper.Dump(empPaolo));

                empMarco.Employees.Add(empPaolo);
                db.SubmitChanges();
                Console.WriteLine(Helper.Dump(empMarco));
                Console.WriteLine(Helper.Dump(empPaolo));
            }
        }
        public static void CircularReferenceAutomatic() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new CircularReferenceNorthwind(Connections.ConnectionString);
                db.Log = Console.Out;
                Employee empMarco = new Employee();
                db.Employees.InsertOnSubmit(empMarco);
                empMarco.FirstName = "Marco";
                empMarco.LastName = "Russo";
                Employee empPaolo = new Employee();
                empPaolo.FirstName = "Paolo";
                empPaolo.LastName = "Pialorsi";
                empPaolo.Employees.Add(empMarco);
                empMarco.Employees.Add(empPaolo);
                db.SubmitChanges();
                Console.WriteLine(Helper.Dump(empMarco));
                Console.WriteLine(Helper.Dump(empPaolo));
            }
        }
        public static void UpdateUsingStoredProcedures() {
            // TransactionScope is necessary to cancel the update before it is committed.
            using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                Northwind db = new Northwind(Connections.ConnectionString);
                db.Log = Console.Out;

                Product product = db.Products.Single(p => p.ProductID == 1);
                product.UnitsInStock += 50;
                db.SubmitChanges();
            }
        }
        public static void ConcurrentUpdates() {
            // TransactionScope is necessary to cancel the update before it is committed.
            // Use of transactions with concurrent updates requires DTC
            // using (TransactionScope ts = new TransactionScope()) {
                Console.WriteLine("*****************");
                ManualResetEvent db1Completed = new ManualResetEvent(false);

                Northwind db1 = new Northwind(Connections.ConnectionString);
                var customer1 = db1.Customers.Single(c => c.CustomerID == "FRANS");
                customer1.ContactName = "Marco Russo";
                customer1.City = "Milano";
                ThreadPool.QueueUserWorkItem(
                    delegate(object state) {
                        Thread.Sleep(1000); // Wait 1 second
                        db1.SubmitChanges();
                        db1Completed.Set();
                    });

                // ------------------------------------------------------
                //
                // BEGIN Listing 6-7
                //

                Northwind db2 = new Northwind(Connections.ConnectionString);
                var customer2 = db2.Customers.Single(c => c.CustomerID == "FRANS");
                customer2.ContactName = "Paolo Pialorsi";
                db1Completed.WaitOne();

                for (int retry = 1; retry < 4; retry++) {
                    Console.WriteLine("Retry loop {0}", retry);

                    try {
                        db2.SubmitChanges(); // Throws exception
                        break;               // Exit from while if submit succeed
                    }
                    catch (ChangeConflictException ex) {
                        Console.WriteLine(ex.Message);
                        DisplayChangeConflict(db2);
                        db2.Refresh(RefreshMode.OverwriteCurrentValues, customer2);
                        //original = db2.Customers.GetOriginalEntityState(customer2);
                        //Console.WriteLine(customer2.ContactName);
                        //Console.WriteLine(customer2.City);
                    }
                }
                //
                // END    Listing 6-7
                //
                // ------------------------------------------------------

                // Restore original value
                Northwind dbRestore = new Northwind(Connections.ConnectionString);
                var cRestore = dbRestore.Customers.Single(c => c.CustomerID == "FRANS");
                cRestore.ContactName = "Paolo Accorti";
                cRestore.City = "Torino";
                dbRestore.SubmitChanges();
            // }
        }

        // ------------------------------------------------------
        //
        // BEGIN Listing 6-7
        //
        private static void DisplayChangeConflict(Northwind db) {
            foreach (ObjectChangeConflict occ in db.ChangeConflicts) {
                MetaTable metatable = db.Mapping.GetTable(occ.Object.GetType());
                Customer entityInConflict = occ.Object as Customer;

                Console.WriteLine("Table={0}, IsResolved={1}", metatable.TableName, occ.IsResolved);
                foreach (MemberChangeConflict mcc in occ.MemberConflicts) {
                    object currVal = mcc.CurrentValue;
                    object origVal = mcc.OriginalValue;
                    object databaseVal = mcc.DatabaseValue;
                    MemberInfo mi = mcc.Member;
                    Console.WriteLine("Member: {0}", mi.Name);
                    Console.WriteLine("current value: {0}", currVal);
                    Console.WriteLine("original value: {0}", origVal);
                    Console.WriteLine("database value: {0}", databaseVal);
                }
            }
        }
        //
        // END    Listing 6-7
        //
        // ------------------------------------------------------

        public static void Transaction() {
            //
            // Listing 6-8
            //
            Northwind db = new Northwind(Connections.ConnectionString);
            Order_Detail orderDetail = db.Order_Details.Single(o => o.OrderID == 10248 && o.ProductID == 42);

            if (orderDetail.Quantity > 10) {
                orderDetail.Discount = 0.05F;
            }

            using (TransactionScope ts = new TransactionScope()) {
                    db.SubmitChanges();
                // Uncomment the following line to commit the transaction
                // ts.Complete();
            }
        }

        public static void TransactionDirect() {
            //
            // Listing 6-9
            //
            Northwind db = new Northwind(Connections.ConnectionString);
            db.Connection.Open();
            Order_Detail orderDetail = db.Order_Details.Single(o => o.OrderID == 10248 && o.ProductID == 42);

            if (orderDetail.Quantity >= 10) {
                orderDetail.Discount = 0.05F;
            }

            using (db.Transaction = db.Connection.BeginTransaction()) {
                db.SubmitChanges();

                // Uncomment the following line to commit the transaction
                //db.Transaction.Commit();
            }
        }

        public static void Serialization() {
            //
            // Listing 6-10
            //
            Northwind db = new Northwind(Connections.ConnectionString);
            var customer = db.Customers.Single(c => c.CustomerID == "WHITC");

            DataContractSerializer dcs = new DataContractSerializer(typeof(Customer));

            StringBuilder sb = new StringBuilder();
            using (XmlWriter writer = XmlWriter.Create(sb)) {
                dcs.WriteObject(writer, customer);
            }
            string xml = sb.ToString();
            Console.WriteLine(xml);

            // Change the entity replacing a string in XML!
            xml = xml.Replace("Karl Jablonski", "John Smith");

            using (TransactionScope ts = new TransactionScope()) {
                //
                // Listing 6-13
                //
                Northwind nw = new Northwind(Connections.ConnectionString);
                nw.Log = Console.Out;

                Customer deserializedCustomer;
                using (XmlTextReader reader = new XmlTextReader(new StringReader(xml))) {
                    deserializedCustomer = (Customer) dcs.ReadObject(reader, true);
                }
                nw.Customers.Attach(deserializedCustomer, customer);

                Console.WriteLine(
                    "ContactName Original={0}, Updated={1}\n", 
                    customer.ContactName, 
                    deserializedCustomer.ContactName);

                nw.SubmitChanges();
            }
        }

        public static void Inheritance() {
            InherithanceNorthwindDataContext nw = new InherithanceNorthwindDataContext(Connections.ConnectionString);
            DevLeap.Linq.LinqToSql.Inheritance.Customer customer = new DevLeap.Linq.LinqToSql.Inheritance.Customer();
            DevLeap.Linq.LinqToSql.Inheritance.Contact contact = new DevLeap.Linq.LinqToSql.Inheritance.SupplierContact();
            customer.Dump();
            contact.Dump();
        }
    }
    //
    // Listing 6-4
    //
    public class CustomNorthwind : Northwind {
        public CustomNorthwind(string connectionString) :
            base(connectionString) { }

        public override void SubmitChanges(ConflictMode failureMode) {
            Helper.DumpChanges(this.GetChangeSet());
            base.SubmitChanges(failureMode);
        }
    }

    //
    // Listing 6-5
    //
    public class CircularReferenceNorthwind : Northwind {
        public CircularReferenceNorthwind(string connectionString) :
            base(connectionString) { }

        public override void SubmitChanges(ConflictMode failureMode) {
            Dictionary<Employee, Employee> employeeReferences;
            employeeReferences = new Dictionary<Employee, Employee>();

            // Remove and save references to other employees
            ChangeSet cs = this.GetChangeSet();
            foreach (object entity in cs.Inserts) {
                if (entity is Employee) {
                    Employee e = (Employee) entity;
                    employeeReferences.Add(e, e.Employee1);
                    e.Employee1 = null;
                }
            }
            // Save Employees without references to other employees
            base.SubmitChanges(failureMode);

            // Restore references to other employees
            foreach (var item in employeeReferences) {
                item.Key.Employee1 = item.Value;
            }
            // Update Employees with references to other employees
            base.SubmitChanges(failureMode);
        }
    }

    //
    // Listing 6-6
    //
    public partial class Northwind {
        partial void UpdateProduct(Product current) {
            Product original = ((Product) (Products.GetOriginalEntityState(current)));

            // Execute the stored procedure for UnitsInStock update
            if (original.UnitsInStock != current.UnitsInStock) {
                int rowCount = this.ExecuteCommand(
                                "exec UpdateProductStock " +
                                "@id={0}, @originalUnits={1}, @decrement={2}",
                                original.ProductID,
                                original.UnitsInStock,
                                (original.UnitsInStock - current.UnitsInStock));
                if (rowCount < 1) {
                    throw new ChangeConflictException();
                }
            }
        }
    }

    public partial class Northwind {
        partial void InsertEmployee(Employee employee) {
            // Commented because following properties does not exists in current implementation
            //employee.CreatedByUser = GetCurrentUser();
            //employee.CreationTime = DateTime.Now;
            Console.WriteLine("Insert Employee : {0}", employee.LastName);
            this.ExecuteDynamicInsert(employee);
        }
    }

/*    public partial class Northwind {
        [Function(Name = "UpdateProduct", IsComposable = false)]
        public IExecuteResult SQL_InsertProduct(string param1) {
            IExecuteResult executeResult =
                this.ExecuteMethodCall(
                         this,
                         (MethodInfo) (MethodInfo.GetCurrentMethod()),
                         param1);
            ISingleResult<CustomerInfo> result =
                (ISingleResult<CustomerInfo>) executeResult.ReturnValue;
            return result;
        }

        partial void InsertProduct(Product instance) {
            this.
            // Execute the stored procedure for UnitsInStock update
            int rowCount = this.ExecuteCommand(
                            "exec UpdateProductStock " +
                            "@id={0}, @originalUnits={1}, @decrement={2}",
                            instance.ProductID,
                            instance.UnitsInStock,
                            (instance.UnitsInStock - instance.UnitsInStock));
            if (rowCount < 1) {
                throw new OptimisticConcurrencyException();
            }
        }
    }
 */
    /*
    public partial class MyDataContext : DataContext {
        public override void SubmitChanges(ConflictMode mode) {
            ChangeSet cs = this.GetChangeSet();
            foreach(object entity in cs.Inserts) {
                if (entity is Employee) {
                    Employee e = (Employee)entity;
                    e.CreatedByUser = GetCurrentUser();
                    e.CreationTime = DateTime.Now;
                }
            }
            base.SubmitChanges(mode);
        }
    }
    */
    /*
    public partial class Northwind : System.Data.Linq.DataContext {
        partial void InsertCustomer(Customer instance) {
            Console.WriteLine("Insert Customer {0}", instance.CustomerID);
            this.ExecuteDynamicInsert(instance);
        }
        partial void UpdateOrder(Order instance) {
            Console.WriteLine("Update Order {0}", instance.OrderID);
            this.ExecuteDynamicUpdate(instance);
        }
    }
     * */

}
